<?php
class mysqlfs {
	public $group=0;
	public $user=NULL;
	protected $security_settings=array();
	protected $connection;
	protected $table;
	private $valid_user_function;
	/**
	 * void __construct(  resource $connction [, string $table = 'index' ] )
	 * Startup function for the class. Registers MySQL connection and table settings
	 */
	function __construct($conn,$tbl='index') {
		$this->connection=$conn;
		$this->table=$tbl;
		$this->valid_user_function=create_function('$user','return false;');
	}
	/**
	 * void setuser ( string $user, string $group )
	 * Sets $user and $group
	 */
	function setuser($user=NULL,$group=0) {
		$this->user=$user;
		$this->group=$group;
	}
	/**
	 * void valid_user ( string $function )
	 * Uses $function to check if a user name is valid for chown
	 */
	function valid_user($function) {
		$this->valid_user_function=create_function('$user',$function);
	}
	/**
	 * array|NULL readdir_full( string $path )
	 * Reads the directory into an array, like scandir()
	 * Returns NULL on failure or
	 * array (
	 *		[type]=> string(1) "f"|"d",
	 *		[filename]=> string ,
	 *		[time]=> int timestamp,
	 *		[rights]=> int(3) UNIX mod,
	 *		[mime]=> string mimetype,
	 *		[size]=> int filesize in bytes,
	 *		[owner]=> string|NULL UNIX owner,
	 *		[group]=> int(1) UNIX owner group
	 *	)
	 * on success
	 */
	 
	 
	function readdir_full($path){
		if(!$this->is_dir($path)) { trigger_error("MySQLfs: $path is not a directory",E_USER_WARNING); return NULL; }
		if(!$this->is_readable($path)) { trigger_error("MySQLfs: $path is not readable",E_USER_WARNING); return NULL; }
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$result=$this->connection->query("SELECT `type`,`filename` , `time` , `rights`, `size` , `owner` , `group` FROM `{$this->table}` WHERE `path` = '$path' AND `filename`!=''");
		$out=array();
		$out[]=$this->getinfo($path.'/..'); $out[0]['filename']='..';
		while($row=$result->fetch_assoc()) {
			if(!$this->is_readable($path.'/'.$row['filename'])) continue;
			$row['time']=strtotime($row['time']);
			$out[]=$row;
		}
		$result->free();
		return $out;
	}
	/**
	 * string Sizetohuman( int $size )
	 * Converts bytes to Normal file sizes
	 */
	function sizetohuman($size) {
      $sizes = array(" Bytes", " KB", " MB", " GB", " TB", " PB", " EB", " ZB", " YB");
      if ($size == 0) { return('n/a'); } else {
      return (round($size/pow(1024, ($i = floor(log($size, 1024)))), 2) . $sizes[$i]); }
	}
	/**
	 * mixed getinfo( string $path [, string $what = NULL] )
	 * Gets system info for a path
	 * Returns the requested info $what or 
	 * array (
	 *		[type]=> string(1) "f"|"d",
	 *		[filename]=> string ,
	 *		[time]=> int timestamp,
	 *		[rights]=> int(3) UNIX mod,
	 *		[mime]=> string mimetype,
	 *		[size]=> int filesize in bytes,
	 *		[mfi]=> int MasterFileIndex,
	 *		[owner]=> string|NULL UNIX owner,
	 *		[group]=> int(1) UNIX owner group
	 *	)
	 */
	function getinfo($path,$what=NULL) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$filename=basename($path);
		$dir=dirname($path);
		if($dir=="\\"||$dir=='.') $dir="";
		$result=$this->connection->query("SELECT `type` , `filename` , `time` , `rights` , `mime` , `size`, `mfi`, `owner` , `group` FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'");
		$info=$result->fetch_assoc();
		$result->free();
		$info['time']=strtotime($info['time']);
		if($info['type']=="d") unset($info['mime']); //Dir's have no mimetype
		if($what!==NULL) $info=$info[$what];
		return $info;
	}
	/**
	 * int filegroup ( string $path )
	 * Gets the file group. The group ID is returned in numerical format
	 */
	function filegroup ($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return (int)$this->getinfo($path,'group');
	}
	/**
	 * int filemtime( string $path )
	 * This function returns the time when the data blocks of a file were being written to, that is, the time when the content of the file was changed.
	 */
	function filemtime($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'time');
	}
	/**
	 * string fileowner ( string $path )
	 * Gets the file owner. 
	 */
	function fileowner($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'owner');
	}
	/**
	 * int fileperms( string $path )
	 * Gets permissions for the given file. 
	 */
	function fileperms($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'rights');
	}
	/** 
	 * int filesize (string $path )
	 * Gets the size for the given file in bytes
	 */
	function filesize($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'size');
	}
	/**
	 * string filetype( string $path )
	 * Returns the type of the given file. 
	 */
	function filetype($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return ($this->getinfo($path,'type')=="f"?'file':'directory');
	}
	/**
	 * string simplify_path( string $path )
	 * Removes /./ and resolves /../
	 */
	function simplify_path($path) {
		if($path=='.'||$path=='\\') $path='';
		if(!preg_match('@^([\w\s/\\.()-])*$@',$path)) {  return NULL; }
        $path = str_replace(array('/', '\\'), '/', $path);
        $parts = array_filter(explode('/', $path), 'strlen');
        $absolutes = array();
        foreach ($parts as $part) {
            if ('.' == $part) continue;
            if ('..' == $part) {
                array_pop($absolutes);
            } else {
                $absolutes[] = $part;
            }
        }
        return implode('/', $absolutes);
    }
	/**
	 * bool is_file ( string $path )
	 * Returns true if the path is a file, false otherwise
	 */
	function is_file($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return ($this->getinfo($path,'type')=="f");
	}
	/**
	 * bool is_dir( string $path )
	 * Returns true if the path is a directory, false otherwise
	 */
	function is_dir($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return ($this->getinfo($path,'type')=="d");
	}
	/**
	 * string|false file_get_contents( string $path )
	 * Returns contents of the file at $path or false if $path isn't a file
	 */
	function file_get_contents($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { trigger_error("MySQLfs: $path is not readable",E_USER_WARNING); return false; }
		if(!$this->is_file($path)) { trigger_error("MySQLfs: $path is not a file",E_USER_WARNING); return false; }
		$filename=basename($path);
		$dir=dirname($path);
		if($dir=='.'||$dir=="\\") $dir="";
		$result=$this->connection->query("SELECT `mfi` FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'")->fetch_assoc();
		return $this->bin64K_getFile($result['mfi']);
	}
	/*
	 * array file ( string $path , int $flags ) 
	 * Reads an entire file into an array. 
	 */
	function file($path,$flags=0) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { trigger_error("MySQLfs: $path is not readable",E_USER_WARNING); return false; }
		if(!$this->is_file($path)) { trigger_error("MySQLfs: $path is not a file",E_USER_WARNING); return false; }
		$c=$this->file_get_contents($path);
		$c=str_replace("\r\n","\n",$c);
		$c=str_replace("\r","\n",$c);
		if($flags==(4||6)) {
			$c=preg_replace("/\n+/","\n",$c);
			$c=preg_replace("/\n$/","",$c);
		}
		$arr=explode("\n",$c);
		if($flags!=2&&$flags!=6) {
			foreach($arr as &$v) $v.="\n";
		}
		return $arr;
		
	}
	/** 
	 * bool is_empty( string $path )
	 * Checks if a directory is empty
	 * Returns true if $path is a file or if $path is an empty directory,
	 * false if the directory $path is not empty
	 */
	function is_empty($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { trigger_error("MySQLfs: $path is not readable",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) return true; //File doesn't contain files
		$dir=dirname($path);
		$result=$this->connection->query("SELECT `rights` FROM `{$this->table}` WHERE `path`='{$path}'");
		$r=$result->num_rows;
		$result->free();
		return ($r==0);
	}		
	/**
	 * bool rmdir( string $path )
	 * Removes the empty directory $path
	 * Returns true on success and false on failure (probably dir not empty)
	 */
	function rmdir($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { trigger_error("MySQLfs: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) { trigger_error("MySQLfs: '$path' is not a directory",E_USER_WARNING); return false; }
		if(!$this->is_empty($path)) { trigger_error("MySQLfs: '$path' is not empty",E_USER_WARNING); return false; }
		$filename=basename($path);
		$dir=dirname($path);
		if($dir=="."||$dir=="\\") $dir='';
		return $this->connection->query("DELETE FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'");
	}
	/**
	 * bool unlink( string $path )
	 * Removes the file at $path
	 * Returns true on success and false on failure
	 */
	function unlink($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { trigger_error("MySQLfs: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->is_file($path)) { trigger_error("MySQLfs: '$path' is not a file",E_USER_WARNING); return false; }
		$filename=basename($path);
		$dir=dirname($path);
		if($dir=="."||$dir=="\\") $dir='';
		$this->bin64K_delete($this->getInfo($path,'mfi'));
		return $this->connection->query("DELETE FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'");
	}
	/**
	 * bool file_exists ( string $path )
	 * Checks if a file exists
	 * Returns true if $path exists, false if $path doesn't exist
	 */
	function file_exists($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$filename=basename($path);
		$dir=dirname($path);
		if($dir=="."||$dir=="\\") $dir='';
		$result=$this->connection->query("SELECT `rights` FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'");
		$r=$result->num_rows;
		$result->free();
		return ($r==1);
	}
	/**
	 * bool mkdir( string $path, [int $mode=NULL] )
	 * Creates a directory $path
	 * if $mode is NULL, it takes the mode of the directory above
	 * Returns true on success or false on failure
	 */
	function mkdir($path,$mode=NULL) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if($this->file_exists($path)) { trigger_error("MySQLfs: $path already exists",E_USER_WARNING); return false; } //File exists
		$filename=basename($path);
		$dir=dirname($path);
		if(!$this->is_writable($dir)) { trigger_error("MySQLfs: $dir is not writable",E_USER_WARNING); return false; }
		if(!$this->is_dir($dir)) { trigger_error("MySQLfs: '$dir' does not exist",E_USER_WARNING); return false; } //Dir above doesn't exist
		if($dir=="."||$dir=="\\") $dir='';
		if($mode===NULL) $mode=$this->getinfo($dir,'rights'); //Get mode from .. dir
		return $this->connection->query("INSERT INTO `{$this->table}` VALUES ('$dir', '$filename', '$mode', '{$this->user}', '{$this->group}', CURRENT_TIMESTAMP, 'd', 'application/octet-stream',0,'')");
	}
	/**
	 * bool mkfile (string $path, [int $mode=NULL, [string $mime=NULL]] )
	 * Creates a file $path
	 * if $mode is NULL, it takes the mode of the directory. If $mime is NULL, it searches the type based on the extention
	 * Returns true on success or false on failure
	 */
	 function mkfile($path,$mode=NULL,$mime=NULL) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if($this->file_exists($path)) { trigger_error("MySQLfs: $path already exists",E_USER_WARNING); return false; } //File exists
		$filename=basename($path);
		$dir=dirname($path);
		if(!$this->is_writable($dir)) { trigger_error("MySQLfs: $dir is not writable",E_USER_WARNING); return false; }
		if($dir=="."||$dir=="\\") $dir='';
		if(!$this->is_dir($dir)) { trigger_error("MySQLfs: '$dir' does not exist",E_USER_WARNING); return false; } //Dir above doesn't exist
		if($mode===NULL) $mode=$this->getinfo($dir,'rights'); // Get mode from .. dir
		if($mime===NULL) $mime=$this->mime_content_type($path); // Get mime from extention
		return $this->connection->query("INSERT INTO `{$this->table}` VALUES ('$dir', '$filename', '$mode', '{$this->user}', '{$this->group}', CURRENT_TIMESTAMP, 'f', '$mime',0,'')");	
	}
	/**
	 * string mime_content_type( string $path );
	 * Gets the content type based on the extention
	 * Returns the mimetype
	 */
	function mime_content_type($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$mime_types = array("323" => "text/h323","acx" => "application/internet-property-stream","ai" => "application/postscript","aif" => "audio/x-aiff","aifc" => "audio/x-aiff","aiff" => "audio/x-aiff","asf" => "video/x-ms-asf","asr" => "video/x-ms-asf","asx" => "video/x-ms-asf","au" => "audio/basic","avi" => "video/x-msvideo","axs" => "application/olescript","bas" => "text/plain","bcpio" => "application/x-bcpio","bin" => "application/octet-stream","bmp" => "image/bmp","c" => "text/plain","cat" => "application/vnd.ms-pkiseccat","cdf" => "application/x-cdf","cer" => "application/x-x509-ca-cert","class" => "application/octet-stream","clp" => "application/x-msclip","cmx" => "image/x-cmx","cod" => "image/cis-cod","cpio" => "application/x-cpio","crd" => "application/x-mscardfile","crl" => "application/pkix-crl","crt" => "application/x-x509-ca-cert","csh" => "application/x-csh","css" => "text/css","dcr" => "application/x-director","der" => "application/x-x509-ca-cert","dir" => "application/x-director","dll" => "application/x-msdownload","dms" => "application/octet-stream","doc" => "application/msword","dot" => "application/msword","dvi" => "application/x-dvi","dxr" => "application/x-director","eps" => "application/postscript","etx" => "text/x-setext","evy" => "application/envoy","exe" => "application/octet-stream","fif" => "application/fractals","flr" => "x-world/x-vrml","gif" => "image/gif","gtar" => "application/x-gtar","gz" => "application/x-gzip","h" => "text/plain","hdf" => "application/x-hdf","hlp" => "application/winhlp","hqx" => "application/mac-binhex40","hta" => "application/hta","htc" => "text/x-component","htm" => "text/html","html" => "text/html","htt" => "text/webviewhtml","ico" => "image/x-icon","ief" => "image/ief","iii" => "application/x-iphone","ins" => "application/x-internet-signup","isp" => "application/x-internet-signup","jfif" => "image/pipeg","jpe" => "image/jpeg","jpeg" => "image/jpeg","jpg" => "image/jpeg","js" => "application/x-javascript","latex" => "application/x-latex","lha" => "application/octet-stream","lsf" => "video/x-la-asf","lsx" => "video/x-la-asf","lzh" => "application/octet-stream","m13" => "application/x-msmediaview","m14" => "application/x-msmediaview","m3u" => "audio/x-mpegurl","man" => "application/x-troff-man","mdb" => "application/x-msaccess","me" => "application/x-troff-me","mht" => "message/rfc822","mhtml" => "message/rfc822","mid" => "audio/mid","mny" => "application/x-msmoney","mov" => "video/quicktime","movie" => "video/x-sgi-movie","mp2" => "video/mpeg","mp3" => "audio/mpeg","mpa" => "video/mpeg","mpe" => "video/mpeg","mpeg" => "video/mpeg","mpg" => "video/mpeg","mpp" => "application/vnd.ms-project","mpv2" => "video/mpeg","ms" => "application/x-troff-ms","mvb" => "application/x-msmediaview","nws" => "message/rfc822","oda" => "application/oda","p10" => "application/pkcs10","p12" => "application/x-pkcs12","p7b" => "application/x-pkcs7-certificates","p7c" => "application/x-pkcs7-mime","p7m" => "application/x-pkcs7-mime","p7r" => "application/x-pkcs7-certreqresp","p7s" => "application/x-pkcs7-signature","pbm" => "image/x-portable-bitmap","pdf" => "application/pdf","pfx" => "application/x-pkcs12","pgm" => "image/x-portable-graymap","pko" => "application/ynd.ms-pkipko","pma" => "application/x-perfmon","pmc" => "application/x-perfmon","pml" => "application/x-perfmon","pmr" => "application/x-perfmon","pmw" => "application/x-perfmon","pnm" => "image/x-portable-anymap","pot" => "application/vnd.ms-powerpoint","ppm" => "image/x-portable-pixmap","pps" => "application/vnd.ms-powerpoint","ppt" => "application/vnd.ms-powerpoint","prf" => "application/pics-rules","ps" => "application/postscript","pub" => "application/x-mspublisher","qt" => "video/quicktime","ra" => "audio/x-pn-realaudio","ram" => "audio/x-pn-realaudio","ras" => "image/x-cmu-raster","rgb" => "image/x-rgb","rmi" => "audio/mid","roff" => "application/x-troff","rtf" => "application/rtf","rtx" => "text/richtext","scd" => "application/x-msschedule","sct" => "text/scriptlet","setpay" => "application/set-payment-initiation","setreg" => "application/set-registration-initiation","sh" => "application/x-sh","shar" => "application/x-shar","sit" => "application/x-stuffit","snd" => "audio/basic","spc" => "application/x-pkcs7-certificates","spl" => "application/futuresplash","src" => "application/x-wais-source","sst" => "application/vnd.ms-pkicertstore","stl" => "application/vnd.ms-pkistl","stm" => "text/html","svg" => "image/svg+xml","sv4cpio" => "application/x-sv4cpio","sv4crc" => "application/x-sv4crc","t" => "application/x-troff","tar" => "application/x-tar","tcl" => "application/x-tcl","tex" => "application/x-tex","texi" => "application/x-texinfo","texinfo" => "application/x-texinfo","tgz" => "application/x-compressed","tif" => "image/tiff","tiff" => "image/tiff","tr" => "application/x-troff","trm" => "application/x-msterminal","tsv" => "text/tab-separated-values","txt" => "text/plain","uls" => "text/iuls","ustar" => "application/x-ustar","vcf" => "text/x-vcard","vrml" => "x-world/x-vrml","wav" => "audio/x-wav","wcm" => "application/vnd.ms-works","wdb" => "application/vnd.ms-works","wks" => "application/vnd.ms-works","wmf" => "application/x-msmetafile","wps" => "application/vnd.ms-works","wri" => "application/x-mswrite","wrl" => "x-world/x-vrml","wrz" => "x-world/x-vrml","xaf" => "x-world/x-vrml","xbm" => "image/x-xbitmap","xla" => "application/vnd.ms-excel","xlc" => "application/vnd.ms-excel","xlm" => "application/vnd.ms-excel","xls" => "application/vnd.ms-excel","xlt" => "application/vnd.ms-excel","xlw" => "application/vnd.ms-excel","xof" => "x-world/x-vrml","xpm" => "image/x-xpixmap","xwd" => "image/x-xwindowdump","z" => "application/x-compress","zip" => "application/zip");
		$filename=basename($path);
		$parts=explode('.',$filename);
		$ext=$parts[count($parts)-1];
		if(!array_key_exists($ext,$mime_types))  return 'application/octet-stream'; //Not in array? Must be default!
		return $mime_types[$ext];
	}
	/**
	 * bool file_put_contents( string $path, string $data )
	 * Puts $data into $pathinfo
	 * Returns true on success or false on failure
	 */
	function file_put_contents($path,$data) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { trigger_error("MySQLfs: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) $this->mkfile($path);
		if(!$this->is_file($path)) { trigger_error("MySQLfs: '$path' is not a file",E_USER_WARNING); return false; }
		$filename=basename($path);
		$dir=dirname($path);
		if($dir=="."||$dir=="\\") $dir='';
		return ($this->bin64K_uploadString($data,$this->getInfo($path,'mfi'))!=0);
	}
	/**
	 * bool is_readable( string $path )
	 * Checks if the file is readable
	 * Returns true is the file is readable, false otherwise
	 */
	function is_readable($path){
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		$rights=$this->getinfo($path,'rights');
		return (($this->getinfo($path,'owner')==$this->user&&$rights[0]>=4)
				||($this->getinfo($path,'group')<=$this->group&&$rights[1]>=4)
				||($rights[2]>=4));
	}
	/**
	 * bool is_writable( string $path )
	 * Checks if the file is writeable
	 * Returns true if the file is writeable, false otherwise
	 */
	function is_writable($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		$rights=$this->getinfo($path,'rights');
		for($i=0;$i<=2;$i++) {
			if($rights[$i]>=4) $rights[$i]=$rights[$i]-4;
		}
		return (($this->getinfo($path,'owner')==$this->user&&$rights[0]>=2)
				||($this->getinfo($path,'group')<=$this->group&&$rights[1]>=2)
				||($rights[2]>=2&&$this->security('wrld_write')));
	}
	function is_writeable($path) {
		return $this->is_writable($path);
	}
	/**
	 * bool is_executable ( string $path )
	 * Checks if the file is executable by php
	 * Returns true if the file is executable, false otherwise
	 */
	function is_executable($path){
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)||!$this->security('exec')) return false;
		$rights=$this->getinfo($path,'rights');
		return (($this->getinfo($path,'owner')==$this->user&&$rights[0]%2==1)
				||((int)$this->getinfo($path,'group')<=$this->group&&$rights[1]%2==1)
				||($rights[2]%2==1));
	}
	 
	/**
	 * bool security ( mixed $key )
	 * You can add some extra security here
	 * To write (once) supply array with key-value pairs
	 * To read, supply the array key
	 */
	function security($key) {
		if(is_array($key)) {
			foreach($key as $k => $v) {
				if(!array_key_exists($k,$this->security_settings)) {
					$this->security_settings[$k]=$v;
				}
				else {
					trigger_error("MySQLfs: Security can only be set once",E_USER_WARNING);
				}
			}
		}
		else {
			if(!array_key_exists($key,$this->security_settings)) return true;
			return $this->security_settings[$key];
		}
	}
	/**
	 * bool chmod ( string $path, int $mode )
	 * UNIX chmod command
	 * Returns true on success, false on failure
	 */
	function chmod($path,$mode) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { trigger_error("MySQLfs: File cannot be written",E_USER_WARNING); return false; }
		$filename=basename($path);
		$dir=dirname($path);
		if($dir=='.'||$dir=='\\') $dir='';
		return $this->connection->query("UPDATE `{$this->table}` SET `rights`='{$this->connection->escape_string($mode)}' WHERE `path` = '$dir' AND `filename` = '$filename'");
	}
	/**
	 * bool chown (string $path, string $user )
	 * UNIX chown command
	 * Only the current owner can chown
	 * Returns true on success, false on failure
	 */
	function chown($path,$user) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if($this->user!=$this->getinfo($path,'owner')) { trigger_error("MySQLfs: Only the current owner can chown the file",E_USER_WARNING); return false; }
		if(!$this->valid_user_function($user)) { trigger_error("MySQLfs: User does not exist",E_USER_WARNING); return false; }
		$filename=basename($path);
		$dir=dirname($path);
		if($dir=='.'||$dir=='\\') $dir='';
		return $this->connection->query("UPDATE `{$this->table}` SET `owner`='{$this->connection->escape_string($user)}' WHERE `path` = '$dir' AND `filename` = '$filename'");
	}
	/**
	 * bool chgrp ( string $path, int $group )
	 * UNIX chgrp command
	 * Only current group and owner can chgrp
	 * Returns true on success, false on failure
	 */
	function chgrp($path,$group) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if($this->user!=$this->getinfo($path,'owner')&&$this->group!=$this->getinfo($path,'group')) { trigger_error("MySQLfs: Only the current group or the current owner can chgrp the file",E_USER_WARNING); return false; }
		$filename=basename($path);
		$dir=dirname($path);
		if($dir=='.'||$dir=='\\') $dir='';
		return $this->connection->query("UPDATE `{$this->table}` SET `group`='{$this->connection->escape_string($group)}' WHERE `path` = '$dir' AND `filename` = '$filename'");
	}
	/**
	 * bool rename ( string $oldpath, string $newpath )
	 * Rename $oldpath to $newpath
	 * You must have reading permissions on $oldpath and writing permissions on $newpath
	 * Returns true on success, false on failure
	 */
	function rename($oldpath,$newpath) {
		if(($oldpath=$this->simplify_path($oldpath))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(($newpath=$this->simplify_path($newpath))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(1)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($oldpath)) { trigger_error("MySQLfs: $oldpath is not readable",E_USER_WARNING); return false; }
		$oldfilename=basename($oldpath);
		$newfilename=basename($newpath);
		$olddir=dirname($oldpath);
		$newdir=dirname($newpath);
		if(!$this->is_dir($newdir)) $newdir=$this->simplify_path($olddir.'/'.$newdir); //Relative?
		if(!$this->is_writable($newdir)) { trigger_error("MySQLfs: $newdir is not writable",E_USER_WARNING); return false; }
		if($this->file_exists($newpath)) { trigger_error("MySQLfs: $newpath already exists",E_USER_WARNING); return false; }
		if(preg_match("|^$oldpath|",$newpath)) { trigger_error("MySQLfs: $newpath is a subdirectory of $oldpath",E_USER_WARNING); return false; }
		if($olddir=='.'||$olddir=='\\') $olddir='';
		if($newdir=='.'||$newdir=='\\') $newdir='';
		$result=$this->connection->query("SELECT `path` FROM `{$this->table}` WHERE `path` LIKE '{$this->simplify_path($olddir.'/'.$oldfilename)}/%' OR `path`='{$this->simplify_path($olddir.'/'.$oldfilename)}'");
		$x=array();
		while($row=$result->fetch_assoc()) {
			$row['newpath']=str_replace($oldpath,$newpath,$row['path']);
			$x[]=$this->connection->query("UPDATE `{$this->table}` SET `path`='$row[newpath]' WHERE `path` = '$row[path]'");
		}
		$result->free();
		$x[]=$this->connection->query("UPDATE `{$this->table}` SET `path`='$newdir',`filename`='$newfilename' WHERE `path` = '$olddir' AND `filename` = '$oldfilename'");
		return !in_array(false,$x);
	}
	/**
	 * bool copy ( string $source, string $dest )
	 * Copy $source to $dest
	 * You must have reading permissions on $source and writing permissions on $dest
	 * Returns true on success, false on failure
	 */	
	function copy($source,$dest) {
		if(($source=$this->simplify_path($source))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(($dest=$this->simplify_path($dest))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(1)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($source)) { trigger_error("MySQLfs: $source is not readable",E_USER_WARNING); return false; }
		$sfilename=basename($source);
		$dfilename=basename($dest);
		$sdir=dirname($source);
		$ddir=dirname($dest);
		if(!$this->is_writable($ddir)) { trigger_error("MySQLfs: $ddir is not writable",E_USER_WARNING); return false; }
		if($this->file_exists($dest)) { trigger_error("MySQLfs: $dest already exists",E_USER_WARNING); return false; }
		if($sdir=='.'||$sdir=='\\') $sdir='';
		if($ddir=='.'||$ddir=='\\') $ddir='';
		$x=array();
		$result=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path` LIKE '$source/%' OR `path` = '$source'");
		while($row=$result->fetch_assoc()) {
			$row['path']=str_replace($source,$dest,$row['path']);
			$x[]=$this->connection->query("INSERT into `{$this->table}` VALUES ('{$row['path']}', '{$row['filename']}','{$row['rights']}','{$row['owner']}','{$row['group']}','{$row['time']}','{$row['type']}','{$row['mime']}','{$row['size']}',NULL)");
			$this->bin64K_uploadString($this->bin64K_getFile($row['mfi']),$this->connection->insert_id);
		}
		$result->free();
		$result=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path` = '$sdir' AND `filename`='$sfilename'");
		$row=$result->fetch_assoc();
		$result->free();
		$row['path']=str_replace($sdir,$ddir,$row['path']);
		$row['filename']=str_replace($sfilename,$dfilename,$row['filename']);
		$x[]=$this->connection->query("INSERT into `{$this->table}` VALUES ('{$row['path']}', '{$row['filename']}','{$row['rights']}','{$row['owner']}','{$row['group']}','{$row['time']}','{$row['type']}','{$row['mime']}','{$row['size']}',NULL)");
		$this->bin64K_uploadString($this->bin64K_getFile($row['mfi']),$this->connection->insert_id);
		return !in_array(false,$x);	
	}
	/**
	 * bool move_uploaded_file ( string $filename, string $destination )
	 * This function checks to ensure that the file designated by filename is a valid upload file (meaning that it was uploaded via PHP's HTTP POST upload mechanism). If the file is valid, it will be moved to the filename given by destination. 
	 * WARNING:
	 * !!! This function works on the actual filesystem and moves it to MySQLfs.
	 * !!! Make sure $filename is a valid name
	 */
	function move_uploaded_file($source,$dest) {
		if(!is_uploaded_file($source)) { trigger_error("MySQLfs: $source is not a valid upload file",E_USER_WARNING); return false; }
		if(($dest=$this->simplify_path($dest))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(1)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($dest)) $this->mkfile($dest);
		return $this->file_put_contents($dest,file_get_contents($source));
	}
	/**
	 * int readfile ( string $path )
	 * Reads a file and writes it to the output buffer. 
	 */
	function readfile($path) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$c=$this->file_get_contents($path);
		echo $c;
		return strlen($c);
	}
	/**
	 * int bin64K_uploadString ( string $data, int $id )
	 * PRIVATE FUNCTION
	 * Method to insert the data into the tables 
	 * creating an entry in the Inodes table 
	 * and the corresponding "chunk" entries in the data table
	 ***
	 * @author Paul Scott
	 * @copyright Paul Scott
	 * @package phpclasses
	 ***
	 * Changes:
	 ** Modified to upload strings
	 ** Modified for MySQLi
	 ** Modified for MySQLfs (only filedata part)
	 ** New encoding: base64
	 */
	private function bin64K_uploadString($data,$id){
		//Update filesize
		$this->connection->query("UPDATE `{$this->table}` SET `size`=".strlen($data)." WHERE `mfi`=$id");
		//Remove old files
		if (!$this->connection->query("DELETE FROM `{$this->table}data` WHERE `masterid`=$id")) {
			trigger_error("MySQLfs: Failure to delete binary inode data!",E_USER_WARNING);
			return 0;
		}
		$data=base64_encode($data); //Base64 encode file
		// Insert into the filedata table
		while (strlen($data)>0) {
			// Make the data mysql insert safe
			$binarydata = substr($data,0,65535);
			if (!$this->connection->query("INSERT INTO `{$this->table}data` values(NULL,$id,'$binarydata')")) 
			{
				trigger_error("MySQLfs: Failure to insert binary inode data row!",E_USER_WARNING);
				return 0;
			}
			$data=substr($data,65535);
		}
		return $id;
	}
	/**
	 * string bin64K_getFile ( int $id )
	 * PRIVATE FUNCTION
	 * Get the splitted file and paste it back together
	 ***
	 * @author Paul Scott
	 * @copyright Paul Scott
	 * @package phpclasses
	 ***
	 * Changes:
	 ** Modified for MySQLi
	 ** Modified for MySQLfs (only filedata part)
	 ** Code performance upgraded: wrap in while block, 1 query
	 ** New encoding: base64
	 */
	private function bin64K_getFile($id) {		
		// Loop through and stream the nodes 1 by 1
		$data='';
		$res=$this->connection->query("SELECT `data` FROM `{$this->table}data` WHERE `masterid` = $id ORDER BY `id`");
		while($block=$res->fetch_assoc()) {
			$data.=$block['data'];
		}
		return base64_decode($data);
	}
	/**
	 * array scandir ( string $path [, int $sorting_order] )
	 * Returns an array of files and directories from the directory. 
	 */
	function scandir($path,$sorting_order=0) {
		if(($path=$this->simplify_path($path))===NULL)  { trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) { trigger_error("MySQLfs: $path does not exist",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) { trigger_error("MySQLfs: $path is not a directory",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { trigger_error("MySQLfs: $path is not readable",E_USER_WARNING); return false; }
		$res=$this->connection->query("SELECT `filename` FROM `{$this->table}` WHERE `path`='$path' ORDER BY `filename` ".(($sorting_order==0)?"ASC":"DESC"));
		$files=array();
		while($f=$res->fetch_array()) {
			$files=$f['filename'];
		}
		return $files;
	}
	private function bin64K_delete($id) {
		return $this->connection->query("DELETE FROM `{$this->table}data` WHERE `masterid` = $id");
	}
}
?>